/*  COSMIX: general n_body simulator, version 0.4 Beta
    Copyright (C) 2001 Shan Mignot

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdlib.h>
#include <math.h>
#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtkgl/gtkglarea.h>
#include <libxml/tree.h>

#include "define.h"
#include "functions.h"

/* The GLU error callback fonction. Causes the program to terminate normally */
void glu_error (GLenum errorCode)
{
  const GLubyte *estring;
  
  estring = gluErrorString (errorCode);
  fprintf (stderr, "OpenGL Error: %s\n", estring);
  exit (0);
}

void redraw_viewport (GtkWidget *widget)
{
  GdkRectangle area;
  
  if (!widget)
    return;

  area.x = 0;
  area.y = 0;
  area.width  = widget->allocation.width;
  area.height = widget->allocation.height;

  gtk_widget_draw (widget, &area);
}

/* Needs to be called for each GtkGLArea to set up the GL context */
void initgl(void)
{
  /* remove back faces */
  glDisable (GL_CULL_FACE);
  glEnable (GL_DEPTH_TEST);
  
  /* speedups */
  glDisable (GL_DITHER);
  glShadeModel (GL_FLAT);
  glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint (GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  /* Geometry */
  glEnable (GL_LINE_STIPPLE);
  glColorMaterial (GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glEnable (GL_COLOR_MATERIAL);  
  glClearColor (.3,.4,.6,1);
}

gint glarea_expose (GtkWidget *widget, GdkEventExpose *event)
{
  GtkGLArea *glarea = GTK_GL_AREA (widget);
  mesh_info *info = (mesh_info*) gtk_object_get_data(GTK_OBJECT(widget),
						     "mesh_info");
  gdouble diagonal, dist, center_x, center_y, center_z;
  
  /* draw only last expose */
  if (event->count > 0)
    return TRUE;

  /* OpenGL calls can be done only if make_current returns true */
  if (gtk_gl_area_make_current(glarea))
    {
      /* basic initialization */
      if (info->do_init == TRUE)
	{
	  initgl();
	  info->do_init = FALSE;
	}

      /* Do not call at each render in practice but only when the mesh is
	 modified
	 if (info->parallel)
	 def_mesh_create (); */

      /* view */
      glMatrixMode (GL_PROJECTION);
      glLoadIdentity ();
      /* The far and near parameters need to be set according to the extent
	 of the geometry, use the bounding_box parameters  */

      diagonal = sqrt ((info->bounding_box[1] - info->bounding_box[0])*
		       (info->bounding_box[1] - info->bounding_box[0]) +
		       (info->bounding_box[3] - info->bounding_box[2])*
		       (info->bounding_box[3] - info->bounding_box[2]) +
		       (info->bounding_box[5] - info->bounding_box[4])*
		       (info->bounding_box[5] - info->bounding_box[4]));
      /* A tolerance factor */
      diagonal += diagonal/10.0;
      diagonal /= 2.0;
      center_x = (info->bounding_box[0] + info->bounding_box[1])/2.0;
      center_y = (info->bounding_box[2] + info->bounding_box[3])/2.0;
      center_z = (info->bounding_box[4] + info->bounding_box[5])/2.0;
      dist = sqrt ((info->lookat[0] - center_x)*
		   (info->lookat[0] - center_x) +
		   (info->lookat[1] - center_y)*
		   (info->lookat[1] - center_y)+
		   (info->lookat[2] - center_z)*
		   (info->lookat[2] - center_z));

      if (info->parallel)
	glOrtho (-diagonal, diagonal, -diagonal, diagonal,
		 dist - diagonal, dist + diagonal);
      else
	/* Make sure the entire trajectories are always visible */
	gluPerspective (info->fov,
			(GLdouble) widget->allocation.width / 
			(GLdouble) widget->allocation.height,
			0.1, dist + diagonal);

      glMatrixMode (GL_MODELVIEW);
      
      /* The background color */
      glClear (GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
      
      /* draw object */
      glLoadIdentity ();
      gluLookAt (info->lookat[0], info->lookat[1], info->lookat[2],
		 info->lookat[3], info->lookat[4], info->lookat[5],
		 info->lookat[6], info->lookat[7], info->lookat[8]);
      
      /* Render the mesh depending on the active notebook page */
      switch (gtk_notebook_get_current_page
	      (GTK_NOTEBOOK (global_widgets[NOTEBOOK].widget)))
	{
	case 0:
	  render_def_mesh ();
	  break;
	case 1:
	  render_grp_mesh ();
	  break;
	default:
	  break;
	}

      /* swap backbuffer to front */
	gtk_gl_area_swapbuffers(glarea);
    }
  
  return TRUE;
}

gint glarea_configure(GtkWidget *widget, GdkEventConfigure *event)
{
  /* OpenGL calls can be done only if make_current returns true */
  if (gtk_gl_area_make_current(GTK_GL_AREA(widget)))
    glViewport(0, 0, widget->allocation.width, widget->allocation.height); 
      
  return TRUE;
}

gint glarea_destroy(GtkWidget *widget)
{
  /* delete mesh info */
  mesh_info *info = (mesh_info*) gtk_object_get_data (GTK_OBJECT(widget),
						      "mesh_info");
  if (info)
    g_free(info);

  return TRUE;
}

gint glarea_button_press (GtkWidget *widget, GdkEventButton *event)
{
  mesh_info *info = (mesh_info*) gtk_object_get_data (GTK_OBJECT(widget),
						      "mesh_info");

  if (event->button == 1)
    {
      /* beginning of drag, reset mouse position */
      info->beginx = event->x;
      info->beginy = event->y;
      return TRUE;
    }

  return FALSE;
}

gint glarea_motion_notify_perspective (GtkWidget *widget, GdkEventMotion *event)
{
  int x, y, i;
  GdkModifierType state;
  mesh_info *info = (mesh_info*) gtk_object_get_data (GTK_OBJECT(widget),
						      "mesh_info");

  if (event->is_hint)
    gdk_window_get_pointer(event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }
  
  if (state & GDK_BUTTON1_MASK)
    {
      /* Rotate around the camera target */
      GLdouble phi = 0.0, theta = 0.0;
      GLdouble vector_u[3], vector_v[3], vector_z[3];
      GLdouble radius, norm, cos_phi, sin_phi, cos_theta, sin_theta;

      phi -= (GLdouble) 0.5*M_PI*(x - info->beginx)/widget->allocation.width;
      theta += (GLdouble) 0.5*M_PI*(y - info->beginy)/widget->allocation.height;

      /* The normed orthogonal basis */
      for (i = 0; i < 3; i++)
	vector_u[i] = info->lookat[i] - info->lookat[3+i];
      radius = (vector_u[0]*vector_u[0] +
		vector_u[1]*vector_u[1] +
		vector_u[2]*vector_u[2]);
      norm = (info->lookat[6]*info->lookat[6] +
	      info->lookat[7]*info->lookat[7] +
	      info->lookat[8]*info->lookat[8]);
      norm = (GLdouble) sqrt (radius/norm);
      for (i = 0; i < 3; i++)
	vector_z[i] = info->lookat[6+i]*norm;
      vector_v[0] = vector_z[1]*vector_u[2] - vector_z[2]*vector_u[1];
      vector_v[1] = vector_z[2]*vector_u[0] - vector_z[0]*vector_u[3];
      vector_v[2] = vector_z[0]*vector_u[1] - vector_z[1]*vector_u[0];
      norm = (vector_v[0]*vector_v[0] +
	      vector_v[1]*vector_v[1] +
	      vector_v[2]*vector_v[2]);
      norm = (GLdouble) sqrt (radius/norm);
      for (i = 0; i < 3; i++)
	vector_v[i] *= norm;

      /* The new camera position and roll */
      cos_phi = (GLdouble) cos (phi);
      sin_phi = (GLdouble) sin (phi);
      cos_theta = (GLdouble) cos (theta);
      sin_theta = (GLdouble) sin (theta);
      for (i = 0; i < 3; i++)
	{
	  info->lookat[i] = (cos_theta*cos_phi*vector_u[i] +
			     cos_theta*sin_phi*vector_v[i] +
			     sin_theta*vector_z[i]) + info->lookat[3+i];
	  info->lookat[6+i] = (-sin_theta*cos_phi*vector_u[i] -
			       sin_theta*sin_phi*vector_v[i] +
			       cos_theta*vector_z[i]);
	}

      /* Orientation has changed, redraw mesh */
      redraw_viewport (widget);
    }

  if (state & GDK_BUTTON2_MASK)
    {
      GLdouble zoom = 1.0;
      /* Zooming drag */
      /* Changes fov rather than dolly the camera
	 info->fov += ((y - info->beginy) / widget->allocation.height) * 40;
	 if (info->fov < DEFAULT_GL_MIN_FOV) info->fov = DEFAULT_GL_MIN_FOV;
	 if (info->fov > DEFAULT_GL_MAX_FOV) info->fov = DEFAULT_GL_MAX_FOV; */
      /* Dolly the camera */
      zoom += (GLdouble) (y - info->beginy) / widget->allocation.height;
      for (i = 0; i < 3; i++)
	info->lookat[i] = info->lookat[i]*zoom + info->lookat[3+i]*(1 - zoom); 

      /* Zoom has changed, redraw mesh */
      redraw_viewport (widget);
    }

  if (state & GDK_BUTTON3_MASK)
    {
      GLdouble vector_u[3], vector_v[3], vector_z[3];
      GLdouble norm;

      /* Use the third button for panning */
      /* The normed orthogonal basis */
      for (i = 0; i < 3; i++)
	vector_u[i] = info->lookat[i] - info->lookat[3+i];
      norm = sqrt (info->lookat[6]*info->lookat[6] +
		   info->lookat[7]*info->lookat[7] +
		   info->lookat[8]*info->lookat[8]);
      for (i = 0; i < 3; i++)
	vector_z[i] = info->lookat[6+i]/norm;
      vector_v[0] = vector_z[1]*vector_u[2] - vector_z[2]*vector_u[1];
      vector_v[1] = vector_z[2]*vector_u[0] - vector_z[0]*vector_u[3];
      vector_v[2] = vector_z[0]*vector_u[1] - vector_z[1]*vector_u[0];
      norm = sqrt (vector_v[0]*vector_v[0] +
		   vector_v[1]*vector_v[1] +
		   vector_v[2]*vector_v[2]);
      for (i = 0; i < 3; i++)
	vector_v[i] /= norm; 
      norm = sqrt (vector_u[0]*vector_u[0] +
		   vector_u[1]*vector_u[1] +
		   vector_u[2]*vector_u[2]);
      norm /= 2;

      /* Vector_u becomes the translation vector applied to both camera
	 and target */
      for (i = 0; i < 3; i++)
	{
	  vector_u[i] = ((-x + info->beginx)/
			 widget->allocation.width*norm*vector_v[i] +
			 (y - info->beginy)/
			 widget->allocation.height*norm*vector_z[i]);
	  info->lookat[i] += vector_u[i];
	  info->lookat[3+i] += vector_u[i];
	}
      
      /* Position has changed, redraw mesh */
      redraw_viewport (widget);
    } 

  info->beginx = x;
  info->beginy = y;
  
  return TRUE;
}

gint glarea_motion_notify_parallel (GtkWidget *widget, GdkEventMotion *event)
{
  return TRUE;
}

void gl_area_set (GtkWidget *frame, GlobalWidgets i)
{
  mesh_info *info;
  GtkWidget *glarea;

  /* Create new OpenGL widget (last argument must be GDK_GL_NONE) so that 
     there exists 2 contexts */
  if (i == DEF_GL_X || i == GRP_GL_X)
    glarea = gtk_gl_area_new_vargs(NULL, GDK_GL_RGBA, GDK_GL_DOUBLEBUFFER,
				   GDK_GL_DEPTH_SIZE, 1, GDK_GL_NONE);
  else 
    {
      if (i == DEF_GL_Y || i == DEF_GL_Z || i == DEF_GL_USER)
	glarea = gtk_gl_area_new_vargs
	  (GTK_GL_AREA (global_widgets[DEF_GL_X].widget),
	   GDK_GL_RGBA, GDK_GL_DOUBLEBUFFER, GDK_GL_DEPTH_SIZE, 1, GDK_GL_NONE);
      else
	{
	  if (i == GRP_GL_Y || i == GRP_GL_Z || i == GRP_GL_USER)
	    glarea = gtk_gl_area_new_vargs
	      (GTK_GL_AREA (global_widgets[GRP_GL_X].widget),
	       GDK_GL_RGBA, GDK_GL_DOUBLEBUFFER, GDK_GL_DEPTH_SIZE,
	       1, GDK_GL_NONE);
	  else return;
	}
    }

  if (glarea == NULL)
    {
      g_print("Can't create GtkGLArea widget\n");
      exit (0);
    } 
  
  /* set up mesh info */
  info = (mesh_info*) g_malloc(sizeof(mesh_info));
  info->do_init = TRUE;
  info->beginx = 0;
  info->beginy = 0;
  info->fov    = 45;
  gtk_object_set_data(GTK_OBJECT(glarea), "mesh_info", info);

  /* The OpenGL viewports are orthogonal for the def page and perspective for
     the grp page */
  if (i == DEF_GL_X || i == DEF_GL_Y || i == DEF_GL_Z || i == DEF_GL_USER)
    info->parallel = TRUE;
  else
    info->parallel = FALSE;

  /* Set up events and signals for OpenGL widgets, handling is different
   according to whether the projection is parallel or perspective. */
  gtk_widget_set_events(glarea,
			GDK_EXPOSURE_MASK|
			GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|
			GDK_POINTER_MOTION_MASK|
			GDK_POINTER_MOTION_HINT_MASK);
  gtk_signal_connect (GTK_OBJECT(glarea), "expose_event",
		      GTK_SIGNAL_FUNC(glarea_expose), NULL);
  if (!info->parallel)
    gtk_signal_connect (GTK_OBJECT(glarea), "motion_notify_event",
			GTK_SIGNAL_FUNC(glarea_motion_notify_perspective), NULL);
  else
    gtk_signal_connect (GTK_OBJECT(glarea), "motion_notify_event",
			  GTK_SIGNAL_FUNC(glarea_motion_notify_parallel), NULL);

  gtk_signal_connect (GTK_OBJECT(glarea), "button_press_event",
		      GTK_SIGNAL_FUNC(glarea_button_press), NULL);
  gtk_signal_connect (GTK_OBJECT(glarea), "configure_event",
		      GTK_SIGNAL_FUNC(glarea_configure), NULL);
  gtk_signal_connect (GTK_OBJECT(glarea), "destroy",
			  GTK_SIGNAL_FUNC (glarea_destroy), NULL);

  gtk_container_add (GTK_CONTAINER (frame), glarea);

  /* destroy this window when exiting from gtk_main(): should this be done 
     here or on the main_window, anyway both produce segfault right now. 
     gtk_quit_add_destroy(1, GTK_OBJECT(glarea)); */
}

void def_mesh_create (void)
{
  GtkWidget *widget;
  GtkCTree *system_tree = NULL;
  GtkCTreeNode *node = NULL, *child = NULL;
  PtObjectKi *ki;
  mesh_info *info = NULL;
  GLfloat center_x, center_y, center_z;
  GLfloat minx = 1e30, maxx = -1e30, miny = 1e30;
  GLfloat maxy = -1e30, minz = 1e30, maxz = -1e30;
  int i;

  if (!global_widgets[SYSTEM_TREE].exists)
    return;
  else system_tree = GTK_CTREE (global_widgets[SYSTEM_TREE].widget);

  /* Scan the system tree for max and min positions */
  node = gtk_ctree_node_nth (system_tree, 0);
  if (node == NULL) return;

  while (node)
    {
      child = GTK_CTREE_ROW (GTK_CTREE_ROW (node)->children)->sibling;
      ki = (PtObjectKi*)
	((Structure*) (gtk_ctree_node_get_row_data (system_tree, child)))
	->structure;
      
      if (ki->position_x_ok && ki->position_y_ok && ki->position_z_ok)
	{
	  if ((GLfloat) ki->positions[X] < minx)
	    minx = (GLfloat) ki->positions[X];
	  if ((GLfloat) ki->positions[X] > maxx)
	    maxx = (GLfloat) ki->positions[X];

	  if ((GLfloat) ki->positions[Y] < miny)
	    miny = (GLfloat) ki->positions[Y];
	  if ((GLfloat) ki->positions[Y] > maxy)
	    maxy = (GLfloat) ki->positions[Y];

	  if ((GLfloat) ki->positions[Z] < minz)
	    minz= (GLfloat) ki->positions[Z];
	  if ((GLfloat) ki->positions[Z] > maxz)
	    maxz = (GLfloat) ki->positions[Z];
	}
      
      node = GTK_CTREE_ROW (node)->sibling;
    }

  for (i = 0; i < 4; i++)
    {
      widget = global_widgets[DEF_GL_X + i].widget; 
      info = (mesh_info*) gtk_object_get_data(GTK_OBJECT (widget), "mesh_info");

      info->bounding_box[0] = minx;
      info->bounding_box[1] = maxx;
      info->bounding_box[2] = miny;
      info->bounding_box[3] = maxy;
      info->bounding_box[4] = minz;
      info->bounding_box[5] = maxz;
    }
  
  /* Setup the viewpoints */
  center_x = (info->bounding_box[1] + info->bounding_box[0])/2.0;
  center_y = (info->bounding_box[3] + info->bounding_box[2])/2.0;
  center_z = (info->bounding_box[5] + info->bounding_box[4])/2.0;

  for (i = 0; i < 4; i++)
    {
      widget = global_widgets[DEF_GL_X + i].widget; 
      info = (mesh_info*) gtk_object_get_data(GTK_OBJECT (widget), "mesh_info");
      
      info->lookat[3] = center_x;
      info->lookat[4] = center_y;
      info->lookat[5] = center_z;

      switch (i)
	{
	case 0:
	  info->lookat[0] = info->bounding_box[1] + 1.0;
	  info->lookat[1] = center_y;
	  info->lookat[2] = center_z;

	  info->lookat[6] = 0.0;
	  info->lookat[7] = 0.0;
	  info->lookat[8] = 1.0;
	  break;
	case 1:
	  info->lookat[0] = center_x;
	  info->lookat[1] = info->bounding_box[3] + 1.0;
	  info->lookat[2] = center_z;
	  
	  info->lookat[6] = 0.0;
	  info->lookat[7] = 0.0;
	  info->lookat[8] = 1.0;
	  break;
	case 2:
	  info->lookat[0] = center_x;
	  info->lookat[1] = center_y;
	  info->lookat[2] = info->bounding_box[5] + 1.0;
	  
	  info->lookat[6] = 0.0;
	  info->lookat[7] = 1.0;
	  info->lookat[8] = 0.0;
	  break;
	case 3:
	  info->lookat[0] = center_x + 1.0;
	  info->lookat[1] = center_y + 1.0;
	  info->lookat[2] = center_z + 1.0;
	  
	  info->lookat[6] = -1.0;
	  info->lookat[7] = -1.0;
	  info->lookat[8] = 2.0;
	  break;
	}
	  
      /* Update the GtkGLArea if not calling at each exposure 
	 if (gtk_gl_area_make_current (GTK_GL_AREA (widget)));
	 {
	 area.width = widget->allocation.width;
 	 area.height = widget->allocation.height;
	 gtk_widget_draw (widget, &area);
	 } */
      //      redraw_viewport (widget);
     }
}

/* The def mesh is under construction so avoid using a display list
   for the display (the mesh is also very simple...) */
void render_def_mesh (void)
{
  GtkCTree *system_tree = NULL;
  GtkCTreeNode *node = NULL, *child = NULL;
  PhyObject *object;
  PtObjectKi *ki;
  gboolean position_ok = FALSE;

  if (!global_widgets[SYSTEM_TREE].exists)
    return;
  else system_tree = GTK_CTREE (global_widgets[SYSTEM_TREE].widget);

  /* Scan the system tree for positions and speeds */
  node = gtk_ctree_node_nth (system_tree, 0);
  if (node == NULL) return;
  
  glPointSize (3.0);
  glBegin (GL_POINTS);

  while (node)
    {
      object = (PhyObject*)
	((Structure*) (gtk_ctree_node_get_row_data (system_tree, node)))
	->structure;
    
      child = GTK_CTREE_ROW (GTK_CTREE_ROW (node)->children)->sibling;
      ki = (PtObjectKi*)
	((Structure*) (gtk_ctree_node_get_row_data (system_tree, child)))
	->structure;

      /* The position has been defined */
      if (ki->position_x_ok && ki->position_y_ok && ki->position_z_ok)
	{
	  position_ok = TRUE;

	  /* Change the color depending on whether the mass is active or not */
	  if (object->active)
	    glColor3f (0.0, 1.0, 0.0);
	  else 
	    glColor3f (1.0, 0.0, 0.0);

	  glVertex3f ((float) ki->positions[X],
		      (float) ki->positions[Y],
		      (float) ki->positions[Z]);
	}
      node = GTK_CTREE_ROW (node)->sibling;
    }

  glEnd ();
  
  if (position_ok)
    {
      node = gtk_ctree_node_nth (system_tree, 0);
      glLineStipple (2, 0xAAAA);
      glBegin(GL_LINES);

      while (node)
	{
	  object = (PhyObject*)
	    ((Structure*) (gtk_ctree_node_get_row_data (system_tree, node)))
	    ->structure;
    
	  child = GTK_CTREE_ROW (GTK_CTREE_ROW (node)->children)->sibling;
	  ki = (PtObjectKi*)
	    ((Structure*) (gtk_ctree_node_get_row_data (system_tree, child)))
	    ->structure;

	  /* The speed has been defined */
	  if (ki->speed_x_ok && ki->speed_y_ok && ki->speed_z_ok)
	    {
	      if (object->active)
		glColor3f (0.0, 1.0, 0.0);
	      else 
		glColor3f (1.0, 0.0, 0.0);
	      glVertex3f ((float) ki->positions[X],
			  (float) ki->positions[Y],
			  (float) ki->positions[Z]);
	      glVertex3f ((float) (ki->positions[X] + ki->speeds[X]),
			  (float) (ki->positions[Y] + ki->speeds[Y]),
			  (float) (ki->positions[Z] + ki->speeds[Z]));
	    }
	  node = GTK_CTREE_ROW (node)->sibling;
	}

      glEnd ();
    }

  glFlush ();
}

/* Creates the mesh to be displayed in the viewports of the grp page.
   Unluckily OpenGL uses "floats" and not "double" floating point
   numbers and arranges them differently. Hence a copy of the simulation
   data needs be created here (only once for each simulation though and
   for all 4 viewports). Also positions the camera so that the whole scene
   is visible within the windows.

   Some (?) optimisation is definitely required. Some normalization might also
   be required to handle uniformely systems of large and small extensions 
   (make sure the system does not go beyond 1e4). */

void grp_mesh_create (void)
{
  int i;
  /*  GLfloat most_ambient[4] = {1.0, 1.0, 1.0, 1.0};
      
      glClear(GL_COLOR_BUFFER_BIT);
      glColor3f(1.0, 1.0, 1.0);
      glLightModelfv(GL_LIGHT_MODEL_AMBIENT, most_ambient);
      glMaterialfv(GL_FRONT, GL_AMBIENT, most_ambient);
      
      glBegin(GL_LINE_STRIP);
      for (i = 0; i <= 30; i++) 
      glEvalCoord1f((GLfloat) i/30.0);
      
      glEnd(); */
  
  /*  if (grp_mesh_list !=0)
      glCallList (grp_mesh_list); */

  /* One display list per trajectory to allow for displaying or hiding each
     of them */
  int object, jx, jy, jz, n_step, n_intervals, n_body;
  gdouble norm1, norm2, dist, max;
  gdouble *x, *y, *z, *vx, *vy, *vz;
  gdouble max_x = 0.0, min_x = 0.0, max_y = 0.0;
  gdouble min_y = 0.0, max_z = 0.0, min_z = 0.0;
  GLUnurbsObj *curve = NULL;
  GLfloat *control_points = NULL, *knots = NULL;
  GLfloat ratio;

  GtkWidget *widget = NULL;
  gdouble width_x, width_y, width_z;
  gdouble center_x, center_y, center_z;
  mesh_info *info = NULL;

  /* The simulation hasn't yet been computed, nothing to render */
  if (!output)
    return;
  
  /* Search for the maximum dimension of the system for normalization */
  n_step = ((NControl*) output->controls)->n_step;
  n_body = output->n_body;
  for (object = 0; object < n_body; object++)
    {
      x = ((NObjectOutput*) output->objects)[object].x;
      y = ((NObjectOutput*) output->objects)[object].y;
      z = ((NObjectOutput*) output->objects)[object].z;

      for (jx = 0; jx < n_step; jx++)
	{
	  if (x[jx] > max_x) max_x = x[jx];
	  if (x[jx] < min_x) min_x = x[jx];
	  if (y[jx] > max_y) max_y = y[jx];
	  if (y[jx] < min_y) min_y = y[jx];
	  if (z[jx] > max_z) max_z = z[jx];
	  if (z[jx] < min_z) min_z = z[jx];
	}
    }
  max = fabs (max_x);
  if (fabs (min_x) > max) max = fabs (min_x);
  if (fabs (max_y) > max) max = fabs (max_y);
  if (fabs (min_y) > max) max = fabs (min_y);
  if (fabs (max_z) > max) max = fabs (max_z);
  if (fabs (min_z) > max) max = fabs (min_z);
  
  /* Set the ratio of the normalization */
  if (max > FRAME_MAX_COORD)
    ratio = (GLfloat) (FRAME_MAX_COORD / max);
  else 
    ratio = 1.0;
  
  if (gtk_gl_area_make_current (GTK_GL_AREA (global_widgets[GRP_GL_X].widget)))
    {
      n_intervals = n_step - 1;
      control_points = (GLfloat*) g_malloc (12*n_intervals*sizeof (GLfloat));
      knots = (GLfloat*) g_malloc (4*n_step*sizeof (GLfloat));

      /* Allocate or reset the various display lists here
	 Defaults to displaying all objects, if display[i] == 0 then 
	 points to the empty display list */
      if (grp_display_list)
	glDeleteLists (grp_display_list->first_list, grp_display_list->range);
      else 
	grp_display_list = g_new0 (GLDisplayLists, 1);
      
      grp_display_list->range = n_body + 1;
      grp_display_list->first_list = glGenLists (grp_display_list->range);
      grp_display_list->display = g_new0 (GLuint, grp_display_list->range);
      for (i = 0; i < n_body + 1; i++)
	grp_display_list->display[i] = i;

      /* The empty display list */
      glNewList (grp_display_list->first_list, GL_COMPILE);
      glEndList ();
      
      /* The knot vector is the same for all trajectories, compute only once */
      jx = 0;
      while (jx < 4*n_step) knots[jx++] = jx / 4;
      
      /* Set the content of the 'object'th display list in the loop */
      max_x = min_x = max_y = min_y = max_z = min_z = 0.0;
      for (object = 0; object < n_body; object++)
	{
	  x = ((NObjectOutput*) output->objects)[object].x;
	  y = ((NObjectOutput*) output->objects)[object].y;
	  z = ((NObjectOutput*) output->objects)[object].z;
	  vx = ((NObjectOutput*) output->objects)[object].vx;
	  vy = ((NObjectOutput*) output->objects)[object].vy;
	  vz = ((NObjectOutput*) output->objects)[object].vz;
	  
	  norm2 = sqrt (vx[0]*vx[0] + vy[0]*vy[0] + vz[0]*vz[0])*ratio;
	  /* The set of control points */
	  for (jx = 0; jx < 12*n_intervals; jx += 3)
	    {
	      int cur = jx/12;
	      int next = cur + 1;
	      
	      jy = jx + 1;
	      jz = jy + 1;
	      
	      switch (jx % 12)
		{
		case 0:
		  dist = sqrt ((x[cur] - x[next])*(x[cur] - x[next]) +
			       (y[cur] - y[next])*(y[cur] - y[next]) +
			       (z[cur] - z[next])*(z[cur] - z[next]))*ratio;
		  norm1 = norm2;
		  control_points[jx] = (GLfloat) x[cur]*ratio;
		  control_points[jy] = (GLfloat) y[cur]*ratio;
		  control_points[jz] = (GLfloat) z[cur]*ratio;
		  break;
		  
		case 3:
		  control_points[jx] = 
		    (GLfloat) (x[cur] + vx[cur]*dist/(3*norm1))*ratio;
		  control_points[jy] = 
		    (GLfloat) (y[cur] + vy[cur]*dist/(3*norm1))*ratio;
		  control_points[jz] = 
		    (GLfloat) (z[cur] + vz[cur]*dist/(3*norm1))*ratio;
		  break;
		  
		case 6:
		  norm2 = sqrt (vx[next]*vx[next] + 
				vy[next]*vy[next] + 
				vz[next]*vz[next])*ratio;
		  control_points[jx] = 
		    (GLfloat) (x[next] - vx[next]*dist/(3*norm2))*ratio;
		  control_points[jy] = 
		    (GLfloat) (y[next] - vy[next]*dist/(3*norm2))*ratio;
		  control_points[jz] = 
		    (GLfloat) (z[next] - vz[next]*dist/(3*norm2))*ratio;
		  break;
		  
		case 9:
		  control_points[jx] = (GLfloat) x[next]*ratio;
		  control_points[jy] = (GLfloat) y[next]*ratio;
		  control_points[jz] = (GLfloat) z[next]*ratio;
		  break;
		}
	    }
	  
	  /* for (i = 0; i < 12*n_intervals; i += 3)
	     {
	     printf ("point %d %d, %f, %f, %f\n", i/3, object, control_points[i],
	     control_points[i + 1], control_points[i + 2]);
	     } 
	 
	     for (i = 0; i < 4*n_step; i++)
	     printf ("knot %d %d: %f\n", i, object, knots[i]); 
	  */
	  
	  glNewList (grp_display_list->first_list + (GLuint) (object + 1),
		     GL_COMPILE);
	  curve = gluNewNurbsRenderer ();
	  gluNurbsCallback (curve, GLU_ERROR, glu_error);
	  
	  gluBeginCurve (curve);
	  glColor3f (1.0, 1.0, 0.0);
	  gluNurbsProperty (curve, GLU_SAMPLING_METHOD, GLU_PARAMETRIC_ERROR);
	  gluNurbsProperty (curve, GLU_PARAMETRIC_TOLERANCE, DEFAULT_GL_ACCURACY);
	  
	  gluNurbsCurve (curve, 4*n_step, knots, 3,
			 &control_points[0], 4, GL_MAP1_VERTEX_3);
	  gluEndCurve (curve);
	  
	  /* Displaying the computed positions */
	  glPointSize(3.0);
	  glColor3f(1.0, 0.0, 0.0);
	  glBegin(GL_POINTS);
	  for (i = 0; i < 12*n_intervals; i += 12) 
	    glVertex3fv (&control_points[i]);
	  /* Add the extremity */
	  glVertex3fv (&control_points[12*n_intervals - 3]);
	  glEnd();
	  
	  glEndList ();
	  /* Free the Nurbs Renderer, is this sufficient or is 
	     additionnal stuff needed to free the ones allocated 
	     within the display list ??? */
	  gluDeleteNurbsRenderer (curve);

	  /* Scan the control_points for the max and min values */
	  for (i = 0; i < 12*n_intervals; i++)
	    {
	      switch (i % 3)
		{
		case 0:
		  if (control_points[i] > max_x) max_x = control_points[i];
		  else
		    {
		      if (control_points[i] < min_x) min_x = control_points[i];
		    }
		  break;
		case 1:
		  if (control_points[i] > max_y) max_y = control_points[i];
		  else
		    {
		      if (control_points[i] < min_y) min_y = control_points[i];
		    }
		  break;
		case 2:
		  if (control_points[i] > max_z) max_z = control_points[i];
		  else
		    {
		      if (control_points[i] < min_z) min_z = control_points[i];
		    }
		  break;
		}
	    }

	  /* Put in the info structure of each GtkGLArea */
	  for (i = 0; i < 4; i++)
	    {
	      info = (mesh_info*) gtk_object_get_data 
		(GTK_OBJECT(global_widgets[GRP_GL_X + i].widget), "mesh_info");
	      info->bounding_box[0] = min_x;
	      info->bounding_box[1] = max_x;
	      info->bounding_box[2] = min_y;
	      info->bounding_box[3] = max_y;
	      info->bounding_box[4] = min_z;
	      info->bounding_box[5] = max_z;
	    }
	}
      
      g_free (control_points);
      g_free (knots);
    }
  
  /* The glListBase needs to be set for each GtkGLArea */
  for (i = 0; i < 4; i++)
    {
      if (gtk_gl_area_make_current
	  (GTK_GL_AREA (global_widgets[GRP_GL_X + i].widget)))
	glListBase (grp_display_list->first_list);
    }
  
  /* Set the initial positions (camera and target) for each GtkGLArea */
  width_x = max_x - min_x;
  width_y = max_y - min_y;
  width_z = max_z - min_z;
  
  /* Put the target at the center of the bounding box */
  center_x = min_x + width_x/2.0;
  center_y = min_y + width_y/2.0;
  center_z = min_z + width_z/2.0;

  for (i = 0; i < 4; i++)
    {
      widget = global_widgets[GRP_GL_X + i].widget;
      info = (mesh_info*) gtk_object_get_data (GTK_OBJECT(widget), "mesh_info");

      if (gtk_gl_area_make_current (GTK_GL_AREA 
				    (global_widgets[GRP_GL_X + i].widget)))
	{
	  switch (i)
	    {
	    case 0:
	      {
		gdouble x_y, x_z, x;

		x_y = 1.1*width_y/(2*tan (M_PI*(info->fov/360.0)));
		x_z = 1.1*width_z/(2*tan (M_PI*(info->fov/360.0)));
		if (x_y > x_z ) x = x_y;
		else x = x_z; 
  		     
		info->lookat[0] = x + max_x;
		info->lookat[1] = center_y;
		info->lookat[2] = center_z;
		info->lookat[3] = center_x;
		info->lookat[4] = center_y;
		info->lookat[5] = center_z;
		info->lookat[6] = 0.0;
		info->lookat[7] = 0.0;
		info->lookat[8] = 1.0;
	      }
	      break;
	    case 1:
	      {
		gdouble y_x, y_z, y;
 
		y_x = 1.1*width_x/(2*tan (M_PI*(info->fov/360.0)));
		y_z = 1.1*width_z/(2*tan (M_PI*(info->fov/360.0)));
		if (y_x > y_z) y = y_x;
		else y = y_z;
  		     
		info->lookat[0] = center_x;
		info->lookat[1] = y + max_y;
		info->lookat[2] = center_z;
		info->lookat[3] = center_x;
		info->lookat[4] = center_y;
		info->lookat[5] = center_z;
		info->lookat[6] = 0.0;
		info->lookat[7] = 0.0;
		info->lookat[8] = 1.0;
	      }	     
	      break;
	    case 2:
	      {
		gdouble z_x, z_y, z;

		z_x = 1.1*width_x/(2*tan (M_PI*(info->fov/360.0)));
		z_y = 1.1*width_y/(2*tan (M_PI*(info->fov/360.0)));
		if (z_x > z_y) z = z_x;
		else z = z_y;
      
		info->lookat[0] = center_x;
		info->lookat[1] = center_y;
		info->lookat[2] = z + max_z;
		info->lookat[3] = center_x;
		info->lookat[4] = center_y;
		info->lookat[5] = center_z;
		info->lookat[6] = 0.0;
		info->lookat[7] = 1.0;
		info->lookat[8] = 0.0;
	      }
	      break;
	    case 3:
	      {
		gdouble user_x, user_y, user_z, user;
      
		user_x = 1.2*width_x/(2*tan (M_PI*(info->fov/360.0)));
		user_y = 1.2*width_y/(2*tan (M_PI*(info->fov/360.0)));
		user_z = 1.2*width_z/(2*tan (M_PI*(info->fov/360.0)));
  		     
		if (user_x > user_y && user_x > user_z)
		  user = user_x;
		else
		  {
		    if (user_y > user_z)
		      user = user_y;
		    else
		      user = user_z;
		  }

		info->lookat[0] = user;
		info->lookat[1] = user;
		info->lookat[2] = user;
		info->lookat[3] = center_x;
		info->lookat[4] = center_y;
		info->lookat[5] = center_z;
		info->lookat[6] = -1.0;
		info->lookat[7] = -1.0;
		info->lookat[8] = 2.0;
	      }
	      break;
	    }

	  /* Update the GtkGLArea */
	  redraw_viewport (widget);
	}
    }
}

/* Uses a display list built right after the simulation has been 
   computed. This is to handle trajectories of arbitrary complexity
   efficiently. Do not use for user-defined drawings as they may change
   (or else one list per drawing).*/
void render_grp_mesh (void)
{
  /* No lists have been set up, nothing to display */
  if (!grp_display_list)
    return;

  /* Is very slow and uses all CPU power when the system has
     a large extension */
  glCallLists (grp_display_list->range, GL_UNSIGNED_INT,
	       grp_display_list->display);
}
